@@ -25,6 +25,8 @@ except ImportError:
 import argcomplete
 import colorama
 from appdirs import user_config_dir
+from halo import Halo
+from spinners.spinners import Spinners
 
 from .ansi import MILCFormatter, ansi_colors, ansi_config, ansi_escape, format_ansi
 from .configuration import Configuration, SubparserWrapper, get_argument_name, handle_store_boolean
@@ -34,7 +36,7 @@ from .attrdict import AttrDict
 class MILC(object):
     """MILC - An Opinionated Batteries Included Framework
     """
-    def __init__(self):
+    def __init__(self, name, version, author):
         """Initialize the MILC object.
         """
         # Setup a lock for thread safety
@@ -42,21 +44,22 @@ class MILC(object):
 
         # Define some basic info
         self.acquire_lock()
+        self.version = version
+        self.author = author
         self._config_store_true = []
         self._config_store_false = []
         self._description = None
         self._entrypoint = None
+        self._spinners = {}
+        self._subcommand = None
         self._inside_context_manager = False
         self.ansi = ansi_colors
         self.arg_only = {}
-        self.config = self.config_source = None
         self.config_file = None
         self.default_arguments = {}
         self.platform = platform()
-
-        # Figure out our program name
-        self.prog_name = sys.argv[0][:-3] if sys.argv[0].endswith('.py') else sys.argv[0]
-        self.prog_name = os.environ.get('MILC_APP_NAME', os.path.basename(self.prog_name))
+        self.prog_name = name
+        self.interactive = sys.stdin.isatty()
         self.release_lock()
 
         # Initialize all the things
@@ -75,7 +78,7 @@ class MILC(object):
     def echo(self, text, *args, **kwargs):
         """Print colorized text to stdout.
 
-        ANSI color strings (such as {fg-blue}) will be converted into ANSI
+        ANSI color strings (such as {fg_blue}) will be converted into ANSI
         escape sequences, and the ANSI reset sequence will be added to all
         strings.
 
@@ -119,12 +122,18 @@ class MILC(object):
         if not capture_output and combined_output:
             raise ValueError("Can't use capture_output=False and combined_output=True at the same time.")
 
-        # On some windows platforms (msys2, possibly others) you have to execute the command through a subshell.
+        # On some windows platforms (msys2, possibly others) you have to
+        # execute the command through a subshell. As well, after execution
+        # stdin is broken so things like milc.questions no longer work.
+        # We pass `stdin=subprocess.DEVNULL` by default to prevent that.
         if 'windows' in self.platform.lower():
             safecmd = map(shlex.quote, command)
             safecmd = ' '.join(safecmd)
             command = [os.environ['SHELL'], '-c', safecmd]
 
+            if 'stdin' not in kwargs:
+                kwargs['stdin'] = subprocess.DEVNULL
+
         # Argument Processing
         if capture_output:
             kwargs['stdout'] = subprocess.PIPE
@@ -133,6 +142,9 @@ class MILC(object):
         if combined_output:
             kwargs['stderr'] = subprocess.STDOUT
 
+        if 'stdin' in kwargs and kwargs['stdin'] is None:
+            del kwargs['stdin']
+
         if text:
             kwargs['universal_newlines'] = True
 
@@ -161,16 +173,16 @@ class MILC(object):
     def print_help(self, *args, **kwargs):
         """Print a help message for the main program or subcommand, depending on context.
         """
-        if self._entrypoint.__name__ in self.subcommands:
-            return self.subcommands[self._entrypoint.__name__].print_help(*args, **kwargs)
+        if self._subcommand:
+            return self.subcommands[self._subcommand.__name__].print_help(*args, **kwargs)
 
         return self._arg_parser.print_help(*args, **kwargs)
 
     def print_usage(self, *args, **kwargs):
         """Print brief description of how the main program or subcommand is invoked, depending on context.
         """
-        if self._entrypoint.__name__ in self.subcommands:
-            return self.subcommands[self._entrypoint.__name__].print_usage(*args, **kwargs)
+        if self._subcommand:
+            return self.subcommands[self._subcommand.__name__].print_usage(*args, **kwargs)
 
         return self._arg_parser.print_usage(*args, **kwargs)
 
@@ -180,11 +192,10 @@ class MILC(object):
         if 'action' in kwargs and kwargs['action'] == 'store_boolean':
             return handle_store_boolean(self, *args, **kwargs)
 
-        completer = None
+        completer = kwargs.get('completer')
 
-        if kwargs.get('completer'):
-            completer = kwargs['completer']
-            del(kwargs['completer'])
+        if 'completer' in kwargs:
+            del kwargs['completer']
 
         self.acquire_lock()
 
@@ -216,7 +227,7 @@ class MILC(object):
         logging.root.setLevel(logging.DEBUG)
         self.release_lock()
 
-        self.add_argument('-V', '--version', version=os.environ.get('MILC_APP_VERSION', 'unknown'), action='version', help='Display the version and exit')
+        self.add_argument('-V', '--version', version=self.version, action='version', help='Display the version and exit')
         self.add_argument('-v', '--verbose', action='store_true', help='Make the logging more verbose')
         self.add_argument('--datetime-fmt', default='%Y-%m-%d %H:%M:%S', help='Format string for datetimes')
         self.add_argument('--log-fmt', default='%(levelname)s %(message)s', help='Format string for printed log output')
@@ -225,6 +236,7 @@ class MILC(object):
         self.add_argument('--log-file', help='File to write log messages to')
         self.add_argument('--color', action='store_boolean', default=ansi_config['color'], help='color in output')
         self.add_argument('--unicode', action='store_boolean', default=ansi_config['unicode'], help='unicode loglevels')
+        self.add_argument('--interactive', action='store_true', help='Force interactive mode even when stdout is not a tty.')
         self.add_argument('--config-file', help='The location for the configuration file')
         self.arg_only['config_file'] = ['general']
 
@@ -254,7 +266,7 @@ class MILC(object):
         if '--config-file' in sys.argv:
             return Path(sys.argv[sys.argv.index('--config-file') + 1]).expanduser().resolve()
 
-        filedir = user_config_dir(appname=self.prog_name, appauthor=os.environ.get('MILC_APP_AUTHOR', self.prog_name.upper()))
+        filedir = user_config_dir(appname=self.prog_name, appauthor=self.author)
         filename = '%s.ini' % self.prog_name
 
         return Path(filedir, filename).resolve()
@@ -266,16 +278,24 @@ class MILC(object):
             raise RuntimeError('You must run this before the with statement!')
 
         def argument_function(handler):
-            subcommand_name = handler.__name__.replace("_", "-")
+            config_name = handler.__name__
+            subcommand_name = config_name.replace("_", "-")
+            arg_name = get_argument_name(self, *args, **kwargs)
 
             if kwargs.get('arg_only'):
-                arg_name = get_argument_name(self, *args, **kwargs)
-
                 if arg_name not in self.arg_only:
                     self.arg_only[arg_name] = []
 
                 self.arg_only[arg_name].append(handler.__name__)
                 del kwargs['arg_only']
+            else:
+                if arg_name not in self.default_arguments:
+                    self.default_arguments[config_name] = {}
+
+                self.default_arguments[config_name][arg_name] = kwargs.get('default')
+
+                if self.config[config_name][arg_name] is None:
+                    self.config[config_name][arg_name] = kwargs.get('default')
 
             if handler is self._entrypoint:
                 self.add_argument(*args, **kwargs)
@@ -304,7 +324,7 @@ class MILC(object):
             self.args[key] = value
 
         if 'entrypoint' in self.args:
-            self._entrypoint = self.args.entrypoint
+            self._subcommand = self.args.entrypoint
 
         self.release_lock()
 
@@ -353,24 +373,17 @@ class MILC(object):
         """Merge CLI arguments into self.config to create the runtime configuration.
         """
         self.acquire_lock()
+        subcommand_name = self._subcommand.__name__ if self._subcommand else None
+
         for argument in self.args:
             if argument in ('subparsers', 'entrypoint'):
                 continue
 
             # Find the argument's section
-            # Underscores in command's names are converted to dashes during initialization.
-            # TODO(Erovia) Find a better solution
-            entrypoint_name = self._entrypoint.__name__.replace("_", "-")
-            if entrypoint_name in self.default_arguments and argument in self.default_arguments[entrypoint_name]:
-                argument_found = True
-                section = self._entrypoint.__name__
             if argument in self.default_arguments['general']:
-                argument_found = True
                 section = 'general'
-
-            if not argument_found:
-                raise RuntimeError('Could not find argument in `self.default_arguments`. This should be impossible!')
-                exit(1)
+            else:
+                section = subcommand_name
 
             if argument not in self.arg_only or section not in self.arg_only[argument]:
                 # Determine the arg value and source
@@ -386,12 +399,11 @@ class MILC(object):
                     passed_on_cmdline = True
                 elif argument in self._config_store_false and not arg_value:
                     passed_on_cmdline = True
-                elif arg_value is not None:
-                    if self.config[section][argument] is None or arg_value != default_value:
-                        passed_on_cmdline = True
+                elif arg_value is not None and (self.config[section][argument] is None or arg_value != default_value):
+                    passed_on_cmdline = True
 
                 # Merge this argument into self.config
-                if passed_on_cmdline and (argument in self.default_arguments['general'] or argument in self.default_arguments[entrypoint_name] or argument not in self.config[entrypoint_name]):
+                if passed_on_cmdline:
                     self.config[section][argument] = arg_value
                     self.config_source[section][argument] = 'argument'
 
@@ -405,10 +417,7 @@ class MILC(object):
         for section_name, section in config.items():
             sane_config.add_section(section_name)
             for option_name, value in section.items():
-                if section_name == 'general':
-                    if option_name in ['config_file']:
-                        continue
-                if value is not None:
+                if self.config_source[section_name][option_name] == 'config_file' and value is not None:
                     sane_config.set(section_name, option_name, str(value))
 
         config_dir = self.config_file.parent
@@ -466,13 +475,15 @@ class MILC(object):
             with self:
                 return self.__call__()
 
-        if not self._entrypoint:
-            raise RuntimeError('No entrypoint provided!')
+        if self._subcommand:
+            return self._subcommand(self)
+        elif self._entrypoint:
+            return self._entrypoint(self)
 
-        return self._entrypoint(self)
+        raise RuntimeError('No entrypoint provided!')
 
     def entrypoint(self, description):
-        """Decorator that marks the entrypoint for simple scripts without subcommands.
+        """Decorator that marks the entrypoint used when a subcommand is not supplied.
         """
         if self._inside_context_manager:
             raise RuntimeError('You must run this before cli()!')
@@ -585,6 +596,10 @@ class MILC(object):
         colorama.init()
         self.parse_args()
         self.merge_args_into_config()
+
+        if self.config.general.interactive:
+            self.interactive = True
+
         self.setup_logging()
 
         return self
@@ -598,3 +613,115 @@ class MILC(object):
             print(exc_type)
             logging.exception(exc_val)
             exit(255)
+
+    def is_spinner(self, name):
+        """Returns true if name is a valid spinner.
+        """
+        return name in Spinners.__members__ or name in self._spinners
+
+    def add_spinner(self, name, spinner):
+        """Adds a new spinner to the list of spinners.
+
+        A spinner is a dictionary with two keys:
+
+            interval
+                An integer that sets how long (in ms) to wait between frames.
+
+            frames
+                A list of frames for this spinner
+        """
+        if self.is_spinner(name):
+            raise ValueError(f'Spinner "{name}" already exists!')
+
+        if not isinstance(spinner, dict):
+            raise ValueError('The spinner must be a dictionary!')
+
+        if 'interval' not in spinner or 'frames' not in spinner:
+            raise ValueError('The spinner must have `interval` and `frames` keys!')
+
+        self._spinners[name] = spinner
+
+    def spinner(self, text, *args, spinner=None, animation='ellipsed', placement='left', color='blue', interval=-1, stream=sys.stdout, enabled=True, **kwargs):
+        """Create a spinner object for showing activity to the user.
+
+        This uses halo <https://github.com/ManrajGrover/halo> behind the scenes, most of the arguments map to Halo objects 1:1.
+
+        There are 3 basic ways to use this:
+
+        * Instantiating a spinner and then using `.start()` and `.stop()` on your object.
+        * Using a context manager (`with cli.spinner(...):`)
+        * Decorate a function (`@cli.spinner(...)`)
+
+        #### Instantiating a spinner
+
+        ```python
+        spinner = cli.spinner(text='Loading', spinner='dots')
+        spinner.start()
+
+        # Do something here
+
+        spinner.stop()
+        ```
+
+        #### Using a context manager
+
+        ```python
+        with cli.spinner(text='Loading', spinner='dots'):
+            # Do something here
+        ```
+
+        #### Decorate a function
+
+        ```python
+        @cli.spinner(text='Loading', spinner='dots')
+        def long_running_function():
+            # Do something here
+        ```
+
+        ### Arguments
+
+            text
+                The text to display next to the spinner. ANSI color strings
+                (such as {fg_blue}) will be converted into ANSI escape
+                sequences, and the ANSI reset sequence will be added to the
+                end of the string.
+
+                If *args or **kwargs are passed they will be used to
+                %-format the text.
+
+            spinner
+                The name of the spinner to use. Available names are here:
+                <https://raw.githubusercontent.com/sindresorhus/cli-spinners/dac4fc6571059bb9e9bc204711e9dfe8f72e5c6f/spinners.json>
+
+            animation
+                The animation to apply to the text if it doesn't fit the
+                terminal. One of `ellipsed`, `bounce`, `marquee`.
+
+            placement
+                Which side of the text to display the spinner on. One of
+                `left`, `right`.
+
+            color
+                Color of the spinner. One of `blue`, `grey`, `red`, `green`,
+                `yellow`, `magenta`, `cyan`, `white`
+
+            interval
+                How long in ms to wait between frames. Defaults to the spinner interval (recommended.)
+
+            stream
+                Stream to write the output. Defaults to sys.stdout.
+
+            enabled
+                Enable or disable the spinner. Defaults to `True`.
+        """
+        if isinstance(spinner, str) and spinner in self._spinners:
+            spinner = self._spinners[spinner]
+
+        return Halo(
+            text=format_ansi(text % (args or kwargs)),
+            spinner=spinner if spinner else 'line',  # FIXME: Grab one of the ascii spinners at random instead of line
+            animation=None if animation == 'ellipsed' else animation,
+            placement=placement,
+            color=color,
+            interval=interval,
+        )
