@@ -1,10 +1,9 @@
-"""Functions to ask the user questions.
-
-These functions can be used to query the user for information.
+"""Sometimes you need to ask the user a question. MILC provides basic functions for collecting and validating user input. You can find these in the `milc.questions` module.
 """
+from getpass import getpass
 
+import milc
 from .ansi import format_ansi
-from . import cli
 
 
 def yesno(prompt, *args, default=None, **kwargs):
@@ -20,19 +19,22 @@ def yesno(prompt, *args, default=None, **kwargs):
     If you add `--yes` and `--no` arguments to your program the user can answer questions by passing command line flags.
 
     ```python
-    @add_argument('-y', '--yes', action='store_true', arg_only=True, help='Answer yes to all questions.')
-    @add_argument('-n', '--no', action='store_true', arg_only=True, help='Answer no to all questions.')
+    @cli.argument('-y', '--yes', action='store_true', arg_only=True, help='Answer yes to all questions.')
+    @cli.argument('-n', '--no', action='store_true', arg_only=True, help='Answer no to all questions.')
     ```
     """
     if not args and kwargs:
         args = kwargs
 
-    if 'no' in cli.args and cli.args.no:
+    if 'no' in milc.cli.args and milc.cli.args.no:
         return False
 
-    if 'yes' in cli.args and cli.args.yes:
+    if 'yes' in milc.cli.args and milc.cli.args.yes:
         return True
 
+    if not milc.cli.interactive:
+        return False
+
     if default is None:
         prompt = prompt + ' [y/n] '
     elif default:
@@ -41,11 +43,9 @@ def yesno(prompt, *args, default=None, **kwargs):
         prompt = prompt + ' [y/N] '
 
     while True:
-        cli.echo('')
         answer = input(format_ansi(prompt % args))
-        cli.echo('')
 
-        if not answer and prompt is not None:
+        if not answer and default is not None:
             return default
 
         elif answer.lower() in ['y', 'yes']:
@@ -55,8 +55,51 @@ def yesno(prompt, *args, default=None, **kwargs):
             return False
 
 
+def password(prompt='Enter password:', *args, confirm=False, confirm_prompt='Confirm password:', confirm_limit=3, validate=None, **kwargs):
+    """Securely receive a password from the user. Returns the password or None.
+
+    | Argument | Description |
+    |----------|-------------|
+    | prompt | The prompt to present to the user. Can include ANSI and format strings like milc's `cli.echo()`. |
+    | confirm | Prompt the user to type the password again and make sure they match. |
+    | confirm_prompt | The prompt to present to the user. Can include ANSI and format strings like milc's `cli.echo()`. |
+    | confirm_limit | Number of attempts to confirm before giving up. Default: 3 |
+    | validate | This is an optional function that can be used to validate the password, EG to check complexity. It should return True or False and have the following signature:<br><br>`def function_name(answer):` |
+    """
+    if not milc.cli.interactive:
+        return None
+
+    if not args and kwargs:
+        args = kwargs
+
+    if prompt[-1] != ' ':
+        prompt += ' '
+
+    if confirm_prompt[-1] != ' ':
+        confirm_prompt += ' '
+
+    i = 0
+    while not confirm_limit or i < confirm_limit:
+        pw = getpass(format_ansi(prompt % args))
+
+        if pw:
+            if validate is not None and not validate(pw):
+                continue
+
+            elif confirm:
+                if getpass(format_ansi(confirm_prompt % args)) == pw:
+                    return pw
+                else:
+                    milc.cli.log.error('Passwords do not match!')
+
+            else:
+                return pw
+
+            i += 1
+
+
 def question(prompt, *args, default=None, confirm=False, answer_type=str, validate=None, **kwargs):
-    """Prompt the user to answer a question with a free-form input.
+    """Allow the user to type in a free-form string to answer.
 
     | Argument | Description |
     |----------|-------------|
@@ -66,16 +109,16 @@ def question(prompt, *args, default=None, confirm=False, answer_type=str, valida
     | answer_type | Specify a type function for the answer. Will re-prompt the user if the function raises any errors. Common choices here include int, float, and decimal.Decimal. |
     | validate | This is an optional function that can be used to validate the answer. It should return True or False and have the following signature:<br><br>`def function_name(answer, *args, **kwargs):` |
     """
-    if not args and kwargs:
-        args = kwargs
+    if not milc.cli.interactive:
+        return default
 
     if default is not None:
         prompt = '%s [%s] ' % (prompt, default)
+    elif prompt and prompt[-1] != ' ':
+        prompt += ' '
 
     while True:
-        cli.echo('')
-        answer = input(format_ansi(prompt % args))
-        cli.echo('')
+        answer = input(format_ansi(prompt % (args or kwargs)))
 
         if answer:
             if validate is not None and not validate(answer, *args, **kwargs):
@@ -86,22 +129,22 @@ def question(prompt, *args, default=None, confirm=False, answer_type=str, valida
                     try:
                         return answer_type(answer)
                     except Exception as e:
-                        cli.log.error('Could not convert answer (%s) to type %s: %s', answer, answer_type.__name__, str(e))
+                        milc.cli.log.error('Could not convert answer (%s) to type %s: %s', answer, answer_type.__name__, str(e))
 
             else:
                 try:
                     return answer_type(answer)
                 except Exception as e:
-                    cli.log.error('Could not convert answer (%s) to type %s: %s', answer, answer_type.__name__, str(e))
+                    milc.cli.log.error('Could not convert answer (%s) to type %s: %s', answer, answer_type.__name__, str(e))
 
         elif default is not None:
             return default
 
 
 def choice(heading, options, *args, default=None, confirm=False, prompt='Please enter your choice: ', **kwargs):
-    """Present the user with a list of options and let them pick one.
+    """Present the user with a list of options and let them select one.
 
-    Returns the value of the item they choose.
+    Users can enter either the number or the text of their choice. This will return the value of the item they choose, not the numerical index.
 
     | Argument | Description |
     |----------|-------------|
@@ -118,20 +161,21 @@ def choice(heading, options, *args, default=None, confirm=False, prompt='Please
     if not args and kwargs:
         args = kwargs
 
+    if not milc.cli.interactive:
+        return default
+
     if prompt and default:
         prompt = prompt + ' [%s] ' % (default + 1,)
+    elif prompt[-1] != ' ':
+        prompt += ' '
 
     while True:
         # Prompt for an answer.
-        cli.echo('')
-        cli.echo(heading % args)
-        cli.echo('')
+        milc.cli.echo(heading % args)
         for i, option in enumerate(options, 1):
-            cli.echo('\t{fg_cyan}%d.{fg_reset} %s', i, option)
+            milc.cli.echo('\t{fg_cyan}%d.{fg_reset} %s', i, option)
 
-        cli.echo('')
         answer = input(format_ansi(prompt))
-        cli.echo('')
 
         # If the user types in one of the options exactly use that
         if answer in options:
@@ -143,14 +187,16 @@ def choice(heading, options, *args, default=None, confirm=False, prompt='Please
         else:
             try:
                 answer = int(answer) - 1
-            except Exception:
-                # Normally we would log the exception here, but in the interest of clean UI we do not.
-                cli.log.error('Invalid choice: %s', answer + 1)
+            except Exception as e:
+                milc.cli.log.error('Invalid choice: %s', answer)
+                milc.cli.log.debug('Could not convert %s to int: %s: %s', answer, e.__class__.__name__, e)
+                if milc.cli.config.general.verbose:
+                    milc.cli.log.exception(e)
                 continue
 
         # Validate the answer
         if answer >= len(options) or answer < 0:
-            cli.log.error('Invalid choice: %s', answer + 1)
+            milc.cli.log.error('Invalid choice: %s', answer + 1)
             continue
 
         if confirm and not yesno('Is the answer "%s" correct?', answer + 1, default=True):
